"""A SignalGroup class that allows connecting to all SignalInstances on the class.

Note that unlike a slot/callback connected to SignalInstance.connect, a slot connected
to SignalGroup.connect does *not* receive the direct arguments that were emitted by a
given SignalInstance. Instead, the slot/callback will receive an EmissionInfo named
tuple, which contains `.signal`: the SignalInstance doing the emitting, and `.args`:
the args that were emitted.

"""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    SupportsIndex,
    overload,
)

from psygnal._signal import _NULL, Signal, SignalInstance, _SignalBlocker

from ._mypyc import mypyc_attr

if TYPE_CHECKING:
    import threading
    from collections.abc import (
        Callable,
        Container,
        Hashable,
        Iterable,
        Iterator,
        Mapping,
    )
    from contextlib import AbstractContextManager

    from psygnal._signal import F, ReducerFunc
    from psygnal._weak_callback import RefErrorChoice, WeakCallback

__all__ = ["EmissionInfo", "SignalGroup"]


@dataclass(slots=True, frozen=True)
class PathStep:
    """A single step in a path to a nested signal.

    This is used to represent a path to a signal that is nested inside an object,
    such as when a signal is emitted from a nested object or a list.  The
    `EmissionInfo.path` is a tuple of `PathStep` objects, where each `PathStep`
    represents either:

    - an _attribute_ access (e.g. `.foo`)
    - an _index_ access (e.g. `[3]`)
    - a _key_ access (e.g. `['user']`)

    !!! note
        _yes, `index` and `key` both just get passed to `__getitem__` in the end, but we
        differentiate them here to make it clearer whether the step is a key in a
        `Mapping` or an index in a `Sequence`._
    """

    attr: str | None = None
    index: SupportsIndex | None = None
    key: Hashable | None = None

    def __post_init__(self) -> None:
        # enforce mutual exclusivity
        if sum(x is not None for x in (self.attr, self.index, self.key)) != 1:
            raise ValueError(
                "PathStep must have exactly one of attr, index, or key set."
            )

    # nice debug display: .foo, [3], ['user']
    def __repr__(self) -> str:  # pragma: no cover
        if self.attr is not None:
            return f".{self.attr}"

        if self.index is not None:
            return f"[{int(self.index)}]"

        key = repr(self.key)
        if len(key) > 20 and not isinstance(self.key, str):
            key = f"{key[:17]}..."
        return f"[{key}]"


@dataclass(slots=True, frozen=True)
class EmissionInfo:
    """Tuple containing information about an emission event.

    Attributes
    ----------
    signal : SignalInstance
        The SignalInstance doing the emitting
    args: tuple
        The args that were emitted
    path: tuple[PathStep, ...]
        A tuple of PathStep objects that describe the path to the signal that was
        emitted. This is useful for nested signals, where the path can be used to
        determine the hierarchy of the signals.  For example, if a signal is emitted
        from a nested object like `bar.foo[3]['user']`, the path will contain the steps
        to get to that object, such as (PathStep(attr='foo'), PathStep(index=3),
        PathStep(key='user')).
    """

    signal: SignalInstance
    args: tuple[Any, ...]
    path: tuple[PathStep, ...] = ()

    def insert_path(self, *path: PathStep) -> EmissionInfo:
        """Return a new EmissionInfo with the given path steps inserted at the front."""
        return EmissionInfo(self.signal, self.args, (*path, *self.path))

    def __post_init__(self) -> None:
        if self.path and not all(
            isinstance(p, (PathStep, str, int)) for p in self.path
        ):
            raise TypeError(
                "EmissionInfo.path must be a tuple of PathStep objects, "
                "or a tuple of str or int."
            )

    def __iter__(self) -> Iterator[Any]:  # pragma: no cover
        """Iterate over the EmissionInfo and all nested EmissionInfos."""
        warnings.warn(
            "`EmissionInfo.__iter__` is no longer a NamedTuple and should not be "
            "iterated over.  Use direct attribute access instead.",
            RuntimeWarning,
            stacklevel=2,
        )
        yield self.signal
        yield self.args


class SignalRelay(SignalInstance):
    """Special SignalInstance that can be used to connect to all signals in a group.

    This class will rarely be instantiated by a user (or anything other than a
    SignalGroup).

    Parameters
    ----------
    signals : Mapping[str, SignalInstance]
        A mapping of signal names to SignalInstance instances.
    instance : Any, optional
        An object to which this `SignalRelay` is bound, by default None
    """

    def __init__(
        self, signals: Mapping[str, SignalInstance], instance: Any = None
    ) -> None:
        super().__init__(signature=(EmissionInfo,), instance=instance)
        self._signals = MappingProxyType(signals)
        self._sig_was_blocked: dict[str, bool] = {}
        self._on_first_connect_callbacks: list[Callable[[], None]] = []

    def _append_slot(self, slot: WeakCallback) -> None:
        super()._append_slot(slot)
        if len(self._slots) == 1:
            self._connect_relay()

    def _connect_relay(self) -> None:
        # Call any registered callbacks on first connection
        for callback in self._on_first_connect_callbacks:
            callback()

        # silence any warnings about failed weakrefs (will occur in compiled version)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for sig in self._signals.values():
                sig.connect(
                    self._slot_relay, check_nargs=False, check_types=False, unique=True
                )

    def _remove_slot(self, slot: int | WeakCallback | Literal["all"]) -> None:
        super()._remove_slot(slot)
        if not self._slots:
            self._disconnect_relay()

    def _disconnect_relay(self) -> None:
        for sig in self._signals.values():
            sig.disconnect(self._slot_relay)

    def _slot_relay(self, *args: Any, loc: PathStep | None = None) -> None:
        """Relay signals from child to parent.

        This is an important method for SignalGroups.  It is the method that is
        responsible for relaying signals from all child signals within the group
        to any connected slots on the group itself.
        """
        emitter = Signal.current_emitter()
        if emitter is None:
            return

        par_path = (loc,) if loc is not None else ()
        # If child already gave us an EmissionInfo, use it;
        if args and isinstance((info := args[0]), EmissionInfo):
            child_info = info.insert_path(*par_path)
        # else wrap its args unchanged
        else:
            child_info = emitter._psygnal_relocate_info_(
                EmissionInfo(signal=emitter, args=args, path=par_path)
            )

        self._run_emit_loop((child_info,))

    def _relay_partial(self, loc: PathStep | None) -> _relay_partial:
        """Return special partial that will call _slot_relay with 'loc'."""
        return _relay_partial(self, loc)

    def connect_direct(
        self,
        slot: Callable | None = None,
        *,
        check_nargs: bool | None = None,
        check_types: bool | None = None,
        unique: bool | Literal["raise"] = False,
        max_args: int | None = None,
    ) -> Callable[[Callable], Callable] | Callable:
        """Connect `slot` to be called whenever *any* Signal in this group is emitted.

        Params are the same as `psygnal.SignalInstance.connect`.  It's probably
        best to check whether all signals are uniform (i.e. have the same signature).

        Parameters
        ----------
        slot : Callable
            A callable to connect to this signal.  If the callable accepts less
            arguments than the signature of this slot, then they will be discarded when
            calling the slot.
        check_nargs : bool | None
            If `True` and the provided `slot` requires more positional arguments than
            the signature of this Signal, raise `TypeError`. by default `True`.
        check_types : bool | None
            If `True`, An additional check will be performed to make sure that types
            declared in the slot signature are compatible with the signature
            declared by this signal, by default `False`.
        unique : bool | Literal["raise"]
            If `True`, returns without connecting if the slot has already been
            connected.  If the literal string "raise" is passed to `unique`, then a
            `ValueError` will be raised if the slot is already connected.
            By default `False`.
        max_args : int, optional
            If provided, `slot` will be called with no more more than `max_args` when
            this SignalInstance is emitted.  (regardless of how many arguments are
            emitted).

        Returns
        -------
        Union[Callable[[Callable], Callable], Callable]
            [description]
        """

        def _inner(slot: Callable) -> Callable:
            for sig in self._signals.values():
                sig.connect(
                    slot,
                    check_nargs=check_nargs,
                    check_types=check_types,
                    unique=unique,
                    max_args=max_args,
                )
            return slot

        return _inner if slot is None else _inner(slot)

    def block(self, exclude: Container[str | SignalInstance] = ()) -> None:
        """Block this signal and all emitters from emitting."""
        super().block()
        for name, sig in self._signals.items():
            if name in exclude or sig in exclude:
                continue
            self._sig_was_blocked[name] = sig._is_blocked
            sig.block()

    def unblock(self) -> None:
        """Unblock this signal and all emitters, allowing them to emit."""
        super().unblock()
        for name, sig in self._signals.items():
            if not self._sig_was_blocked.pop(name, False):
                sig.unblock()

    def blocked(
        self, exclude: Container[str | SignalInstance] = ()
    ) -> AbstractContextManager[None]:
        """Context manager to temporarily block all emitters in this group.

        Parameters
        ----------
        exclude : iterable of str or SignalInstance, optional
            An iterable of signal instances or names to exempt from the block,
            by default ()
        """
        return _SignalBlocker(self, exclude=exclude)

    def disconnect(self, slot: Callable | None = None, missing_ok: bool = True) -> None:
        """Disconnect slot from all signals.

        Parameters
        ----------
        slot : callable, optional
            The specific slot to disconnect.  If `None`, all slots will be disconnected,
            by default `None`
        missing_ok : bool, optional
            If `False` and the provided `slot` is not connected, raises `ValueError.
            by default `True`

        Raises
        ------
        ValueError
            If `slot` is not connected and `missing_ok` is False.
        """
        for sig in self._signals.values():
            sig.disconnect(slot, missing_ok)
        super().disconnect(slot, missing_ok)

    def _slot_index(self, slot: Callable) -> int:
        """Get index of `slot` in `self._slots`. Return -1 if not connected.

        In interpreted mode, `_relay_partial` callbacks are stored as
        weak references to the callable object itself, so comparing the
        dereferenced callback to the provided `_relay_partial` works. In compiled
        mode (mypyc), `weak_callback` may normalize a `_relay_partial` to a strong
        reference to its `__call__` method (a MethodWrapperType), to avoid segfaults on
        garbage collection. In that case the default WeakCallback equality logic is the
        correct and more robust path.

        Therefore, try the base implementation first (which compares normalized
        WeakCallback objects). If that fails and we're dealing with a
        `_relay_partial`, fall back to comparing the dereferenced callable to the
        provided slot for backward compatibility.
        """
        # First, try the standard equality path used by SignalInstance
        idx = super()._slot_index(slot)
        if idx != -1:
            return idx

        # Fallback: handle direct comparison for `_relay_partial` instances
        if isinstance(slot, _relay_partial):
            with self._lock:
                for i, s in enumerate(self._slots):
                    deref = s.dereference()
                    # Case 1: stored deref is the _relay_partial itself (interpreted)
                    if deref == slot:
                        return i
                    # Case 2: compiled path where we stored __call__ bound method
                    # (these will never hit on codecov, but they are covered in tests)
                    owner = getattr(deref, "__self__", None)  # pragma: no cover
                    if (
                        isinstance(owner, _relay_partial) and owner == slot
                    ):  # pragma: no cover
                        return i
        return -1  # pragma: no cover


# NOTE
# To developers. Avoid adding public names to this class, as it is intended to be
# a container for user-determined names.  If names must be added, try to prefix
# with "psygnal_" to avoid conflicts with user-defined names.
@mypyc_attr(allow_interpreted_subclasses=True)
class SignalGroup:
    """A collection of signals that can be connected to as a single unit.

    This class is not intended to be instantiated directly.  Instead, it should be
    subclassed, and the subclass should define Signal instances as class attributes.
    The SignalGroup will then automatically collect these signals and provide a
    SignalRelay instance (at `group.all`) that can be used to connect to all of the
    signals in the group.

    This class is used in both the EventedModels and the evented dataclass patterns.
    See also: `psygnal.SignalGroupDescriptor`, which provides convenient and explicit
    way to create a SignalGroup on a dataclass-like class.

    Parameters
    ----------
    instance : Any, optional
        An object to which this `SignalGroup` is bound, by default None

    Attributes
    ----------
    all : SignalRelay
        A special SignalRelay instance that can be used to connect to all signals in
        this group.

    Examples
    --------
    ```python
    from psygnal import Signal, SignalGroup


    class MySignals(SignalGroup):
        sig1 = Signal()
        sig2 = Signal()


    group = MySignals()
    group.all.connect(print)  # connect to all signals in the group

    list(group)  # ['sig1', 'sig2']
    len(group)  # 2
    group.sig1 is group["sig1"]  # True
    ```
    """

    _psygnal_signals: ClassVar[Mapping[str, Signal]]
    _psygnal_uniform: ClassVar[bool] = False
    _psygnal_name_conflicts: ClassVar[set[str]]
    _psygnal_aliases: ClassVar[dict[str, str | None]]

    _psygnal_instances: dict[str, SignalInstance]

    def __init__(self, instance: Any = None) -> None:
        cls = type(self)
        if not hasattr(cls, "_psygnal_signals"):
            raise TypeError(
                "Cannot instantiate `SignalGroup` directly.  Use a subclass instead."
            )

        self._psygnal_instances = {
            name: (
                sig._create_signal_instance(self)
                if name in cls._psygnal_name_conflicts
                else sig.__get__(self, cls)
            )
            for name, sig in cls._psygnal_signals.items()
        }
        self._psygnal_relay = SignalRelay(self._psygnal_instances, instance)

    def __init_subclass__(
        cls,
        strict: bool = False,
        signal_aliases: Mapping[str, str | None] = {},
    ) -> None:
        """Collects all Signal instances on the class under `cls._psygnal_signals`."""
        # Collect Signals and remove from class attributes
        # Use dir(cls) instead of cls.__dict__ to get attributes from super()
        forbidden = {
            k for k in getattr(cls, "__dict__", ()) if k.startswith("_psygnal")
        }
        if forbidden:
            raise TypeError(
                f"SignalGroup subclass cannot have attributes starting with '_psygnal'."
                f" Found: {forbidden}"
            )

        _psygnal_signals = {}
        for k in dir(cls):
            val = getattr(cls, k, None)
            if isinstance(val, Signal):
                _psygnal_signals[k] = val

        # Collect the Signals also from super-class
        # When subclassing, the Signals have been removed from the attributes,
        # look for cls._psygnal_signals also
        cls._psygnal_signals = {
            **getattr(cls, "_psygnal_signals", {}),
            **_psygnal_signals,
        }

        # Emit warning for signal names conflicting with SignalGroup attributes
        reserved = set(dir(SignalGroup))
        cls._psygnal_name_conflicts = conflicts = {
            k
            for k in cls._psygnal_signals
            if k in reserved or k.startswith(("_psygnal", "psygnal"))
        }
        if conflicts:
            for name in conflicts:
                if isinstance(getattr(cls, name), Signal):
                    delattr(cls, name)
            Names = "Names" if len(conflicts) > 1 else "Name"
            Are = "are" if len(conflicts) > 1 else "is"
            warnings.warn(
                f"{Names} {sorted(conflicts)!r} {Are} reserved. You cannot use these "
                "names to access SignalInstances as attributes on a SignalGroup. (You "
                "may still access them as keys to __getitem__: `group['name']`).",
                UserWarning,
                stacklevel=2,
            )

        aliases = getattr(cls, "_psygnal_aliases", {})
        cls._psygnal_aliases = {**aliases, **signal_aliases}
        cls._psygnal_uniform = _is_uniform(cls._psygnal_signals.values())
        if strict and not cls._psygnal_uniform:
            raise TypeError(
                "All Signals in a strict SignalGroup must have the same signature"
            )
        super().__init_subclass__()

    @property
    def instance(self) -> Any:
        """Object that owns this `SignalGroup`."""
        return self._psygnal_relay.instance

    @property
    def all(self) -> SignalRelay:
        """SignalInstance that can be used to connect to all signals in this group.

        Examples
        --------
        ```python
        from psygnal import Signal, SignalGroup


        class MySignals(SignalGroup):
            sig1 = Signal()
            sig2 = Signal()


        group = MySignals()
        group.sig2.connect(...)  # connect to a single signal by name
        group.all.connect(...)  # connect to all signals in the group
        ```
        """
        return self._psygnal_relay

    @property
    def signals(self) -> Mapping[str, SignalInstance]:
        """A mapping of signal names to SignalInstance instances."""
        # TODO: deprecate this property
        # warnings.warn(
        #     "Accessing the `signals` property on a SignalGroup is deprecated. "
        #     "Use __iter__ to iterate over all signal names, and __getitem__ or "
        #     "getattr to access signal instances. This will be an error in a future.",
        #     FutureWarning,
        #     stacklevel=2,
        # )
        return MappingProxyType(self._psygnal_instances)

    def __len__(self) -> int:
        """Return the number of signals in the group (not including the relay)."""
        return len(self._psygnal_instances)

    def __getitem__(self, item: str) -> SignalInstance:
        """Get a signal instance by name."""
        return self._psygnal_instances[item]

    def __iter__(self) -> Iterator[str]:
        """Yield the names of all signals in the group."""
        return iter(self._psygnal_instances)

    def __contains__(self, item: str) -> bool:
        """Return True if the group contains a signal with the given name."""
        # this is redundant with __iter__ and can be removed, but only after
        # removing the deprecation warning in __getattr__
        return item in self._psygnal_instances

    def __repr__(self) -> str:
        """Return repr(self)."""
        name = self.__class__.__name__

        if self.instance is not None:
            owner_type = type(self.instance).__name__
            owner_repr = f" on <{owner_type} instance at {id(self.instance):#x}>"
        else:
            owner_repr = ""
        return f"<SignalGroup {name!r}{owner_repr} with {len(self)} signals>"

    @classmethod
    def psygnals_uniform(cls) -> bool:
        """Return true if all signals in the group have the same signature."""
        return cls._psygnal_uniform

    @classmethod
    def is_uniform(cls) -> bool:
        """Return true if all signals in the group have the same signature."""
        warnings.warn(
            "The `is_uniform` method on SignalGroup is deprecated. Use "
            "`psygnals_uniform` instead. This will be an error in v0.11.",
            FutureWarning,
            stacklevel=2,
        )
        return cls._psygnal_uniform

    def __deepcopy__(self, memo: dict[int, Any]) -> SignalGroup:
        # TODO:
        # This really isn't a deep copy. Should we also copy connections?
        # a working deepcopy is important for pydantic support, but in most cases
        # it will be a group without any signals connected
        return type(self)(instance=self._psygnal_relay.instance)

    # The rest are passthrough methods to the SignalRelay.
    # The full signatures are here to make mypy and IDEs happy.
    # parity with SignalInstance methods is tested in test_group.py

    @overload
    def connect(
        self,
        *,
        thread: threading.Thread | Literal["main", "current"] | None = ...,
        check_nargs: bool | None = ...,
        check_types: bool | None = ...,
        unique: bool | Literal["raise"] = ...,
        max_args: int | None = None,
        on_ref_error: RefErrorChoice = ...,
        priority: int = ...,
    ) -> Callable[[F], F]: ...

    @overload
    def connect(
        self,
        slot: F,
        *,
        thread: threading.Thread | Literal["main", "current"] | None = ...,
        check_nargs: bool | None = ...,
        check_types: bool | None = ...,
        unique: bool | Literal["raise"] = ...,
        max_args: int | None = None,
        on_ref_error: RefErrorChoice = ...,
        priority: int = ...,
    ) -> F: ...

    def connect(
        self,
        slot: F | None = None,
        *,
        thread: threading.Thread | Literal["main", "current"] | None = None,
        check_nargs: bool | None = None,
        check_types: bool | None = None,
        unique: bool | Literal["raise"] = False,
        max_args: int | None = None,
        on_ref_error: RefErrorChoice = "warn",
        priority: int = 0,
        emit_on_evented_child_events: bool = False,
    ) -> Callable[[F], F] | F:
        if slot is None:
            return self._psygnal_relay.connect(
                thread=thread,
                check_nargs=check_nargs,
                check_types=check_types,
                unique=unique,
                max_args=max_args,
                on_ref_error=on_ref_error,
                priority=priority,
                emit_on_evented_child_events=emit_on_evented_child_events,
            )
        else:
            return self._psygnal_relay.connect(
                slot,
                thread=thread,
                check_nargs=check_nargs,
                check_types=check_types,
                unique=unique,
                max_args=max_args,
                on_ref_error=on_ref_error,
                priority=priority,
                emit_on_evented_child_events=emit_on_evented_child_events,
            )

    def connect_direct(
        self,
        slot: Callable | None = None,
        *,
        check_nargs: bool | None = None,
        check_types: bool | None = None,
        unique: bool | Literal["raise"] = False,
        max_args: int | None = None,
    ) -> Callable[[Callable], Callable] | Callable:
        return self._psygnal_relay.connect_direct(
            slot,
            check_nargs=check_nargs,
            check_types=check_types,
            unique=unique,
            max_args=max_args,
        )

    def disconnect(self, slot: Callable | None = None, missing_ok: bool = True) -> None:
        return self._psygnal_relay.disconnect(slot=slot, missing_ok=missing_ok)

    def block(self, exclude: Container[str | SignalInstance] = ()) -> None:
        return self._psygnal_relay.block(exclude=exclude)

    def unblock(self) -> None:
        return self._psygnal_relay.unblock()

    def blocked(
        self, exclude: Container[str | SignalInstance] = ()
    ) -> AbstractContextManager[None]:
        return self._psygnal_relay.blocked(exclude=exclude)

    def pause(self) -> None:
        return self._psygnal_relay.pause()

    def resume(self, reducer: ReducerFunc | None = None, initial: Any = _NULL) -> None:
        return self._psygnal_relay.resume(reducer=reducer, initial=initial)

    def paused(
        self, reducer: ReducerFunc | None = None, initial: Any = _NULL
    ) -> AbstractContextManager[None]:
        return self._psygnal_relay.paused(reducer=reducer, initial=initial)


def _is_uniform(signals: Iterable[Signal]) -> bool:
    """Return True if all signals have the same signature."""
    seen: set[tuple[str, ...]] = set()
    for s in signals:
        v = tuple(str(p.annotation) for p in s.signature.parameters.values())
        if seen and v not in seen:  # allow zero or one
            return False
        seen.add(v)
    return True


class _relay_partial:
    """Small, single-purpose, mypyc-friendly variant of functools.partial.

    Used to call SignalRelay._slot_relay with a specific loc.
    __hash__ and __eq__ are implemented to make this object hashable and
    comparable to other _relay_partial objects, so that adding it to a set
    twice will not create two separate entries.
    """

    __slots__ = ("loc", "relay")

    def __init__(self, relay: SignalRelay, loc: PathStep | None = None) -> None:
        self.relay = relay
        self.loc = loc

    def __call__(self, *args: Any) -> Any:
        return self.relay._slot_relay(*args, loc=self.loc)

    def __hash__(self) -> int:
        return hash((self.relay, self.loc))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, _relay_partial)
            and self.relay == other.relay
            and self.loc == other.loc
        )



from __future__ import annotations

from threading import Timer
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar

if TYPE_CHECKING:
    import inspect
    from collections.abc import Callable
    from typing import ParamSpec

    Kind = Literal["throttler", "debouncer"]
    EmissionPolicy = Literal["trailing", "leading"]

    P = ParamSpec("P")
else:
    # just so that we don't have to depend on a new version of typing_extensions
    # at runtime
    P = TypeVar("P")


class _ThrottlerBase(Generic[P]):
    _timer: Timer

    def __init__(
        self,
        func: Callable[P, Any],
        interval: int = 100,
        policy: EmissionPolicy = "leading",
    ) -> None:
        self.__wrapped__: Callable[P, Any] = func
        self._interval: float = interval / 1000
        self._policy: EmissionPolicy = policy
        self._has_pending: bool = False
        self._timer: Timer = Timer(0, lambda: None)
        self._timer.start()
        self._args: tuple[Any, ...] = ()
        self._kwargs: dict[str, Any] = {}

        # this mimics what functools.wraps does, but avoids __dict__ usage and other
        # things that won't work with mypyc... HOWEVER, most of these dynamic
        # assignments won't work in mypyc anyway (they just do nothing.)
        self.__module__: str = getattr(func, "__module__", "")
        self.__name__: str = getattr(func, "__name__", "")
        self.__qualname__: str = getattr(func, "__qualname__", "")
        self.__doc__: str | None = getattr(func, "__doc__", None)
        self.__annotations__: dict[str, Any] = getattr(func, "__annotations__", {})

    def _actually_call(self) -> None:
        self._has_pending = False
        self.__wrapped__(*self._args, **self._kwargs)
        self._start_timer()

    def _call_if_has_pending(self) -> None:
        if self._has_pending:
            self._actually_call()

    def _start_timer(self) -> None:
        self._timer.cancel()
        self._timer = Timer(self._interval, self._call_if_has_pending)
        self._timer.start()

    def cancel(self) -> None:
        """Cancel any pending calls."""
        self._has_pending = False
        self._timer.cancel()

    def flush(self) -> None:
        """Force a call if there is one pending."""
        self._call_if_has_pending()

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        raise NotImplementedError("Subclasses must implement this method.")

    @property
    def __signature__(self) -> inspect.Signature:
        import inspect

        return inspect.signature(self.__wrapped__)


class Throttler(_ThrottlerBase, Generic[P]):
    """Class that prevents calling `func` more than once per `interval`.

    Parameters
    ----------
    func : Callable[P, Any]
        a function to wrap
    interval : int, optional
        the minimum interval in ms that must pass before the function is called again,
        by default 100
    policy : EmissionPolicy, optional
        Whether to invoke the function on the "leading" or "trailing" edge of the
        wait timer, by default "leading"
    """

    _timer: Timer

    def __init__(
        self,
        func: Callable[P, Any],
        interval: int = 100,
        policy: EmissionPolicy = "leading",
    ) -> None:
        super().__init__(func, interval, policy)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Call underlying function."""
        self._has_pending = True
        self._args = args
        self._kwargs = kwargs

        if not self._timer.is_alive():
            if self._policy == "leading":
                self._actually_call()
            else:
                self._start_timer()


class Debouncer(_ThrottlerBase, Generic[P]):
    """Class that waits at least `interval` before calling `func`.

    Parameters
    ----------
    func : Callable[P, Any]
        a function to wrap
    interval : int, optional
        the minimum interval in ms that must pass before the function is called again,
        by default 100
    policy : EmissionPolicy, optional
        Whether to invoke the function on the "leading" or "trailing" edge of the
        wait timer, by default "trailing"
    """

    _timer: Timer

    def __init__(
        self,
        func: Callable[P, Any],
        interval: int = 100,
        policy: EmissionPolicy = "trailing",
    ) -> None:
        super().__init__(func, interval, policy)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Call underlying function."""
        self._has_pending = True
        self._args = args
        self._kwargs = kwargs

        if not self._timer.is_alive() and self._policy == "leading":
            self._actually_call()
        self._start_timer()


def throttled(
    func: Callable[P, Any] | None = None,
    timeout: int = 100,
    leading: bool = True,
) -> Throttler[P] | Callable[[Callable[P, Any]], Throttler[P]]:
    """Create a throttled function that invokes func at most once per timeout.

    The throttled function comes with a `cancel` method to cancel delayed func
    invocations and a `flush` method to immediately invoke them. Options
    to indicate whether func should be invoked on the leading and/or trailing
    edge of the wait timeout. The func is invoked with the last arguments provided
    to the throttled function. Subsequent calls to the throttled function return
    the result of the last func invocation.

    This decorator may be used with or without parameters.

    Parameters
    ----------
    func : Callable
        A function to throttle
    timeout : int
        Timeout in milliseconds to wait before allowing another call, by default 100
    leading : bool
        Whether to invoke the function on the leading edge of the wait timer,
        by default True

    Examples
    --------
    ```python
    from psygnal import Signal, throttled

    class MyEmitter:
        changed = Signal(int)

    def on_change(val: int)
        # do something possibly expensive
        ...

    emitter = MyEmitter()

    # connect the `on_change` whenever `emitter.changed` is emitted
    # BUT, no more than once every 50 milliseconds
    emitter.changed.connect(throttled(on_change, timeout=50))
    ```
    """

    def deco(func: Callable[P, Any]) -> Throttler[P]:
        policy: EmissionPolicy = "leading" if leading else "trailing"
        return Throttler(func, timeout, policy)

    return deco(func) if func is not None else deco


def debounced(
    func: Callable[P, Any] | None = None,
    timeout: int = 100,
    leading: bool = False,
) -> Debouncer[P] | Callable[[Callable[P, Any]], Debouncer[P]]:
    """Create a debounced function that delays invoking `func`.

    `func` will not be invoked until `timeout` ms have elapsed since the last time
    the debounced function was invoked.

    The debounced function comes with a `cancel` method to cancel delayed func
    invocations and a `flush` method to immediately invoke them. Options
    indicate whether func should be invoked on the leading and/or trailing edge
    of the wait timeout. The func is invoked with the *last* arguments provided to
    the debounced function. Subsequent calls to the debounced function return the
    result of the last `func` invocation.

    This decorator may be used with or without parameters.

    Parameters
    ----------
    func : Callable
        A function to throttle
    timeout : int
        Timeout in milliseconds to wait before allowing another call, by default 100
    leading : bool
        Whether to invoke the function on the leading edge of the wait timer,
        by default False

    Examples
    --------
    ```python
    from psygnal import Signal, debounced

    class MyEmitter:
        changed = Signal(int)

    def on_change(val: int)
        # do something possibly expensive
        ...

    emitter = MyEmitter()

    # connect the `on_change` whenever `emitter.changed` is emitted
    # ONLY once at least 50 milliseconds have passed since the last signal emission.
    emitter.changed.connect(debounced(on_change, timeout=50))
    ```
    """

    def deco(func: Callable[P, Any]) -> Debouncer[P]:
        policy: EmissionPolicy = "leading" if leading else "trailing"
        return Debouncer(func, timeout, policy)

    return deco(func) if func is not None else deco



"""These utilities may help when using signals and evented objects."""

from __future__ import annotations

from contextlib import contextmanager, suppress
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any
from warnings import warn

from ._group import EmissionInfo, SignalGroup
from ._signal import SignalInstance

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterator

__all__ = ["iter_signal_instances", "monitor_events"]


def _default_event_monitor(info: EmissionInfo) -> None:
    print(f"{info.signal.name}.emit{info.args!r}")


@contextmanager
def monitor_events(
    obj: Any | None = None,
    logger: Callable[[EmissionInfo], Any] = _default_event_monitor,
    include_private_attrs: bool = False,
) -> Iterator[None]:
    """Context manager to print or collect events emitted by SignalInstances on `obj`.

    Parameters
    ----------
    obj : object, optional
        Any object that has an attribute that has a SignalInstance (or SignalGroup).
        If None, all SignalInstances will be monitored.
    logger : Callable[[EmissionInfo], None], optional
        A optional function to handle the logging of the event emission.  This function
        must take two positional args: a signal name string, and a tuple that contains
        the emitted arguments. The default logger simply prints the signal name and
        emitted args.
    include_private_attrs : bool
        Whether private signals (starting with an underscore) should also be logged,
        by default False
    """
    code = getattr(logger, "__code__", None)
    _old_api = bool(code and code.co_argcount > 1)

    if obj is None:
        # install the hook globally
        if _old_api:
            raise ValueError(
                "logger function must take a single argument (an EmissionInfo instance)"
            )
        before, SignalInstance._debug_hook = SignalInstance._debug_hook, logger
    else:
        if _old_api:
            warn(
                "logger functions must now take a single argument (an instance of "
                "psygnal.EmissionInfo). Please update your logger function.",
                stacklevel=2,
            )
        disconnectors = set()
        for siginst in iter_signal_instances(obj, include_private_attrs):
            if _old_api:

                def _report(*args: Any, signal: SignalInstance = siginst) -> None:
                    logger(signal.name, args)  # type: ignore

            else:

                def _report(*args: Any, signal: SignalInstance = siginst) -> None:
                    logger(EmissionInfo(signal, args))

            disconnectors.add(partial(siginst.disconnect, siginst.connect(_report)))

    try:
        yield
    finally:
        if obj is None:
            SignalInstance._debug_hook = before
        else:
            for disconnector in disconnectors:
                disconnector()


def iter_signal_instances(
    obj: Any, include_private_attrs: bool = False
) -> Generator[SignalInstance, None, None]:
    """Yield all `SignalInstance` attributes found on `obj`.

    Parameters
    ----------
    obj : object
        Any object that has an attribute that has a SignalInstance (or SignalGroup).
    include_private_attrs : bool
        Whether private signals (starting with an underscore) should also be logged,
        by default False

    Yields
    ------
    SignalInstance
        SignalInstances (and SignalGroups) found as attributes on `obj`.
    """
    # SignalGroup
    if isinstance(obj, SignalGroup):
        for sig in obj:
            yield obj[sig]
        return

    # Signal attached to Class
    for n in dir(obj):
        if not include_private_attrs and n.startswith("_"):
            continue
        with suppress(Exception):  # if we can't access the attribute, skip it
            attr = getattr(obj, n)
            if isinstance(attr, SignalInstance):
                yield attr
            if isinstance(attr, SignalGroup):
                yield attr._psygnal_relay


_COMPILED_EXTS = (".so", ".pyd")
_BAK = "_BAK"


def decompile() -> None:
    """Mangle names of mypyc-compiled files so that they aren't used.

    This function requires write permissions to the psygnal source directory.
    """
    for suffix in _COMPILED_EXTS:  # pragma: no cover
        for path in Path(__file__).parent.rglob(f"**/*{suffix}"):
            path.rename(path.with_suffix(f"{suffix}{_BAK}"))


def recompile() -> None:
    """Fix all name-mangled mypyc-compiled files so that they ARE used.

    This function requires write permissions to the psygnal source directory.
    """
    for suffix in _COMPILED_EXTS:  # pragma: no cover
        for path in Path(__file__).parent.rglob(f"**/*{suffix}{_BAK}"):
            path.rename(path.with_suffix(suffix))



from collections.abc import Iterable
from importlib.metadata import PackageNotFoundError, PackagePath
from importlib.metadata import files as package_files
from pathlib import Path

try:
    import psygnal

    PSYGNAL_DIR = Path(psygnal.__file__).parent
except ImportError:
    PSYGNAL_DIR = Path(__file__).parent.parent


def binary_files(file_list: Iterable[PackagePath | Path]) -> list[Path]:
    return [Path(file) for file in file_list if file.suffix in {".so", ".pyd"}]


def create_hiddenimports() -> list[str]:
    res = ["queue", "mypy_extensions", "__future__"]

    try:
        files_list = package_files("psygnal")
    except PackageNotFoundError:
        return res

    if files_list is None:
        return res

    modules = binary_files(files_list)

    if len(modules) < 2:
        # This is a workaround for a bug in importlib.metadata in editable mode

        src_path = PSYGNAL_DIR.parent

        modules = [
            x.relative_to(src_path)
            for x in binary_files(PSYGNAL_DIR.iterdir())
            + binary_files(src_path.iterdir())
        ]

    for module in modules:
        res.append(str(module).split(".")[0].replace("/", ".").replace("\\", "."))

    return res


hiddenimports = create_hiddenimports()



from __future__ import annotations

from abc import ABC, abstractmethod
from math import inf
from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from collections.abc import Coroutine
    from typing import Any, Literal, Protocol, TypeAlias

    import anyio.streams.memory
    import trio

    from psygnal._weak_callback import WeakCallback

    SupportedBackend: TypeAlias = Literal["asyncio", "anyio", "trio"]
    QueueItem: TypeAlias = tuple["WeakCallback", tuple[Any, ...]]

    class EventLike(Protocol):
        def is_set(self) -> bool:
            """Return ``True`` if the flag is set, ``False`` if not."""
            ...

        async def wait(self) -> Coroutine | bool | None:
            """Wait until the flag is set."""
            ...


_ASYNC_BACKEND: _AsyncBackend | None = None


def get_async_backend() -> _AsyncBackend | None:
    """Get the current async backend. Returns None if no backend is set."""
    return _ASYNC_BACKEND


def clear_async_backend() -> None:
    """Clear the current async backend. Primarily for testing purposes."""
    global _ASYNC_BACKEND
    if _ASYNC_BACKEND is not None:
        # Cancel any running tasks if it's asyncio and loop is not closed
        if isinstance(_ASYNC_BACKEND, AsyncioBackend):
            _ASYNC_BACKEND.close()
        # Close anyio streams
        elif isinstance(_ASYNC_BACKEND, AnyioBackend):
            _ASYNC_BACKEND.close()
        # Close trio channels
        elif isinstance(_ASYNC_BACKEND, TrioBackend):
            if hasattr(_ASYNC_BACKEND, "_send_channel"):
                _ASYNC_BACKEND._send_channel.close()
            # Note: trio receive channels don't have a close method
    _ASYNC_BACKEND = None


@overload
def set_async_backend(backend: Literal["asyncio"]) -> AsyncioBackend: ...
@overload
def set_async_backend(backend: Literal["anyio"]) -> AnyioBackend: ...
@overload
def set_async_backend(backend: Literal["trio"]) -> TrioBackend: ...
def set_async_backend(backend: SupportedBackend = "asyncio") -> _AsyncBackend:
    """Set the async backend to use. Must be one of: 'asyncio', 'anyio', 'trio'.

    This should be done as early as possible, and *must* be called before calling
    `SignalInstance.connect` with a coroutine function.
    """
    global _ASYNC_BACKEND

    if _ASYNC_BACKEND and _ASYNC_BACKEND._backend != backend:  # pragma: no cover
        # allow setting the same backend multiple times, for tests
        raise RuntimeError(f"Async backend already set to: {_ASYNC_BACKEND._backend}")

    if backend == "asyncio":
        _ASYNC_BACKEND = AsyncioBackend()
    elif backend == "anyio":
        _ASYNC_BACKEND = AnyioBackend()
    elif backend == "trio":
        _ASYNC_BACKEND = TrioBackend()
    else:  # pragma: no cover
        raise RuntimeError(
            f"Async backend not supported: {backend}.  "
            "Must be one of: 'asyncio', 'anyio', 'trio'"
        )

    return _ASYNC_BACKEND


class _AsyncBackend(ABC):
    def __init__(self, backend: str):
        self._backend = backend

    @property
    @abstractmethod
    def running(self) -> EventLike: ...

    @abstractmethod
    def put(self, item: QueueItem) -> None: ...

    @abstractmethod
    async def run(self) -> None: ...

    async def call_back(self, item: QueueItem) -> None:
        cb, args = item
        if func := cb.dereference():
            await func(*args)


class AsyncioBackend(_AsyncBackend):
    def __init__(self) -> None:
        super().__init__("asyncio")
        import asyncio

        self._asyncio = asyncio
        self._queue: asyncio.Queue[tuple] = asyncio.Queue()
        self._task = asyncio.create_task(self.run())
        self._loop = asyncio.get_running_loop()
        self._running = asyncio.Event()

    @property
    def running(self) -> EventLike:
        """Return the event indicating if the backend is running."""
        return self._running

    def put(self, item: QueueItem) -> None:
        self._queue.put_nowait(item)

    def close(self) -> None:
        """Close the asyncio backend and cancel tasks."""
        if hasattr(self, "_task") and not self._task.done():
            self._task.cancel()

    async def run(self) -> None:
        if self._running.is_set():
            return

        self._running.set()
        try:
            while True:
                item = await self._queue.get()
                try:
                    await self.call_back(item)
                except Exception:
                    # Log the exception but continue running
                    # This prevents one bad callback from crashing the backend
                    import traceback

                    traceback.print_exc()
        except self._asyncio.CancelledError:
            pass
        except RuntimeError as e:  # pragma: no cover
            if not self._loop.is_closed():
                raise e
        finally:
            self._running.clear()


class AnyioBackend(_AsyncBackend):
    _send_stream: anyio.streams.memory.MemoryObjectSendStream[QueueItem]
    _receive_stream: anyio.streams.memory.MemoryObjectReceiveStream[QueueItem]

    def __init__(self) -> None:
        super().__init__("anyio")
        import anyio

        self._anyio = anyio
        self._send_stream, self._receive_stream = anyio.create_memory_object_stream(
            max_buffer_size=inf
        )
        self._running = anyio.Event()

    @property
    def running(self) -> EventLike:
        """Return the event indicating if the backend is running."""
        return self._running

    def put(self, item: QueueItem) -> None:
        self._send_stream.send_nowait(item)

    def close(self) -> None:
        """Close the anyio streams."""
        if hasattr(self, "_send_stream"):
            self._send_stream.close()
        if hasattr(self, "_receive_stream"):
            self._receive_stream.close()

    async def run(self) -> None:
        if self._running.is_set():
            return  # pragma: no cover

        self._running.set()
        try:
            async with self._receive_stream:
                async for item in self._receive_stream:
                    try:
                        await self.call_back(item)
                    except Exception:
                        # Log the exception but continue running
                        import traceback

                        traceback.print_exc()
        finally:
            self._running = self._anyio.Event()
            # Ensure streams are closed
            self.close()


class TrioBackend(_AsyncBackend):
    _send_channel: trio._channel.MemorySendChannel[QueueItem]
    _receive_channel: trio.abc.ReceiveChannel[QueueItem]

    def __init__(self) -> None:
        super().__init__("trio")
        import trio

        self._trio = trio
        self._send_channel, self._receive_channel = trio.open_memory_channel(
            max_buffer_size=inf
        )
        self._running = self._trio.Event()

    @property
    def running(self) -> EventLike:
        """Return the event indicating if the backend is running."""
        return self._running

    def put(self, item: tuple) -> None:
        self._send_channel.send_nowait(item)

    async def run(self) -> None:
        if self._running.is_set():
            return  # pragma: no cover

        self._running.set()
        try:
            async for item in self._receive_channel:
                try:
                    await self.call_back(item)
                except Exception:
                    # Log the exception but continue running
                    import traceback

                    traceback.print_exc()
        finally:
            self._running = self._trio.Event()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypeVar, overload

from psygnal._group_descriptor import SignalGroupDescriptor

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping

    from psygnal._group_descriptor import EqOperator, FieldAliasFunc

__all__ = ["evented"]

T = TypeVar("T", bound=type)


@overload
def evented(
    cls: T,
    *,
    events_namespace: str = "events",
    equality_operators: dict[str, EqOperator] | None = None,
    warn_on_no_fields: bool = ...,
    cache_on_instance: bool = ...,
    connect_child_events: bool = ...,
    signal_aliases: Mapping[str, str | None] | FieldAliasFunc | None = ...,
) -> T: ...
@overload
def evented(
    cls: Literal[None] | None = None,
    *,
    events_namespace: str = "events",
    equality_operators: dict[str, EqOperator] | None = None,
    warn_on_no_fields: bool = ...,
    cache_on_instance: bool = ...,
    connect_child_events: bool = ...,
    signal_aliases: Mapping[str, str | None] | FieldAliasFunc | None = ...,
) -> Callable[[T], T]: ...
def evented(
    cls: T | None = None,
    *,
    events_namespace: str = "events",
    equality_operators: dict[str, EqOperator] | None = None,
    warn_on_no_fields: bool = True,
    cache_on_instance: bool = True,
    connect_child_events: bool = True,
    signal_aliases: Mapping[str, str | None] | FieldAliasFunc | None = None,
) -> Callable[[T], T] | T:
    """A decorator to add events to a dataclass.

    See also the documentation for
    [`SignalGroupDescriptor`][psygnal.SignalGroupDescriptor].  This decorator is
    equivalent setting a class variable named `events` to a new
    `SignalGroupDescriptor` instance.

    Note that this decorator will modify `cls` *in place*, as well as return it.

    !!!tip
        It is recommended to use the `SignalGroupDescriptor` descriptor rather than
        the decorator, as it it is more explicit and provides for easier static type
        inference.

    Parameters
    ----------
    cls : type
        The class to decorate.
    events_namespace : str
        The name of the namespace to add the events to, by default `"events"`
    equality_operators : dict[str, Callable] | None
        A dictionary mapping field names to equality operators (a function that takes
        two values and returns `True` if they are equal). These will be used to
        determine if a field has changed when setting a new value.  By default, this
        will use the `__eq__` method of the field type, or np.array_equal, for numpy
        arrays.  But you can provide your own if you want to customize how equality is
        checked. Alternatively, if the class has an `__eq_operators__` class attribute,
        it will be used.
    warn_on_no_fields : bool
        If `True` (the default), a warning will be emitted if no mutable dataclass-like
        fields are found on the object.
    cache_on_instance : bool, optional
        If `True` (the default), a newly-created SignalGroup instance will be cached on
        the instance itself, so that subsequent accesses to the descriptor will return
        the same SignalGroup instance.  This makes for slightly faster subsequent
        access, but means that the owner instance will no longer be pickleable.  If
        `False`, the SignalGroup instance will *still* be cached, but not on the
        instance itself.
    connect_child_events : bool, optional
        If `True`, will connect events from all fields on the dataclass whose type is
        also "evented" (as determined by the `psygnal.is_evented` function,
        which returns True if the class has been decorated with `@evented`, or if it
        has a SignalGroupDescriptor) to the group on the parent object. By default
        True.
        This is useful for nested evented dataclasses, where you want to monitor events
        emitted from arbitrarily deep children on the parent object.
    signal_aliases: Mapping[str, str | None] | Callable[[str], str | None] | None
        If defined, a mapping between field name and signal name. Field names that are
        not `signal_aliases` keys are not aliased (the signal name is the field name).
        If the dict value is None, do not create a signal associated with this field.
        If a callable, the signal name is the output of the function applied to the
        field name. If the output is None, no signal is created for this field.
        If None, defaults to an empty dict, no aliases.
        Default to None

    Returns
    -------
    type
        The decorated class, which gains a new SignalGroup instance at the
        `events_namespace` attribute (by default, `events`).

    Raises
    ------
    TypeError
        If the class is frozen or is not a class.

    Examples
    --------
    ```python
    from psygnal import evented
    from dataclasses import dataclass


    @evented
    @dataclass
    class Person:
        name: str
        age: int = 0
    ```
    """

    def _decorate(cls: T) -> T:
        if not isinstance(cls, type):  # pragma: no cover
            raise TypeError("evented can only be used on classes")
        if any(k.startswith("_psygnal") for k in getattr(cls, "__annotations__", {})):
            raise TypeError("Fields on an evented class cannot start with '_psygnal'")

        descriptor: SignalGroupDescriptor = SignalGroupDescriptor(
            equality_operators=equality_operators,
            warn_on_no_fields=warn_on_no_fields,
            cache_on_instance=cache_on_instance,
            connect_child_events=connect_child_events,
            signal_aliases=signal_aliases,
        )
        # as a decorator, this will have already been called
        descriptor.__set_name__(cls, events_namespace)
        setattr(cls, events_namespace, descriptor)
        return cls

    return _decorate(cls) if cls is not None else _decorate



"""Psygnal implements the observer pattern for Python.

It emulates the signal/slot pattern from Qt, but it does not require Qt.
"""

import os
from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ._evented_model import EventedModel


try:
    __version__ = version("psygnal")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0"
__author__ = "Talley Lambert"
__email__ = "talley.lambert@gmail.com"

__all__ = [
    "EmissionInfo",
    "EmitLoopError",
    "EventedModel",
    "PathStep",
    "Signal",
    "SignalGroup",
    "SignalGroupDescriptor",
    "SignalInstance",
    "__version__",
    "_compiled",
    "debounced",
    "emit_queued",
    "evented",
    "get_async_backend",
    "get_evented_namespace",
    "is_evented",
    "set_async_backend",
    "throttled",
]


if os.getenv("PSYGNAL_UNCOMPILED"):
    import warnings

    warnings.warn(
        "PSYGNAL_UNCOMPILED no longer has any effect. If you wish to run psygnal "
        "without compiled files, you can run:\n\n"
        'python -c "import psygnal.utils; psygnal.utils.decompile()"\n\n'
        "(You will need to reinstall psygnal to get the compiled version back.)",
        stacklevel=2,
    )

from ._async import get_async_backend, set_async_backend
from ._evented_decorator import evented
from ._exceptions import EmitLoopError
from ._group import EmissionInfo, PathStep, SignalGroup
from ._group_descriptor import SignalGroupDescriptor, get_evented_namespace, is_evented
from ._queue import emit_queued
from ._signal import Signal, SignalInstance, _compiled
from ._throttler import debounced, throttled


def __getattr__(name: str) -> Any:  # pragma: no cover
    if name == "EventedModel":
        from ._evented_model import EventedModel

        return EventedModel
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


del os, TYPE_CHECKING



# this pyi file exists until we can use ParamSpec with mypyc in the main file.
from collections.abc import Callable
from typing import Any, Generic, Literal, ParamSpec, overload

P = ParamSpec("P")

Kind = Literal["throttler", "debouncer"]
EmissionPolicy = Literal["trailing", "leading"]

class _ThrottlerBase(Generic[P]):
    def __init__(
        self,
        func: Callable[P, Any],
        interval: int = 100,
        policy: EmissionPolicy = "leading",
    ) -> None: ...
    def cancel(self) -> None:
        """Cancel any pending calls."""

    def flush(self) -> None:
        """Force a call if there is one pending."""

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None: ...

class Throttler(_ThrottlerBase[P]):
    """Class that prevents calling `func` more than once per `interval`.

    Parameters
    ----------
    func : Callable[P, Any]
        a function to wrap
    interval : int, optional
        the minimum interval in ms that must pass before the function is called again,
        by default 100
    policy : EmissionPolicy, optional
        Whether to invoke the function on the "leading" or "trailing" edge of the
        wait timer, by default "leading"
    """

class Debouncer(_ThrottlerBase[P]):
    """Class that waits at least `interval` before calling `func`.

    Parameters
    ----------
    func : Callable[P, Any]
        a function to wrap
    interval : int, optional
        the minimum interval in ms that must pass before the function is called again,
        by default 100
    policy : EmissionPolicy, optional
        Whether to invoke the function on the "leading" or "trailing" edge of the
        wait timer, by default "trailing"
    """

@overload
def throttled(
    func: Callable[P, Any],
    timeout: int = 100,
    leading: bool = True,
) -> Throttler[P]: ...
@overload
def throttled(
    func: Literal[None] | None = None,
    timeout: int = 100,
    leading: bool = True,
) -> Callable[[Callable[P, Any]], Throttler[P]]: ...
@overload
def debounced(
    func: Callable[P, Any],
    timeout: int = 100,
    leading: bool = False,
) -> Debouncer[P]: ...
@overload
def debounced(
    func: Literal[None] | None = None,
    timeout: int = 100,
    leading: bool = False,
) -> Callable[[Callable[P, Any]], Debouncer[P]]: ...
